package edu.lawrence.invasiveandroid

import android.app.Application
import androidx.compose.runtime.mutableStateListOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.toMutableStateList
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.viewModelScope
import androidx.room.Room
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import okhttp3.OkHttpClient
import retrofit2.Retrofit
import retrofit2.converter.gson.GsonConverterFactory
import java.util.concurrent.TimeUnit

class InvasiveModel(application: Application): AndroidViewModel(application) {
    private lateinit var restInterface : InvasiveApiService
    private lateinit var dao : InvasiveDao
    val species = mutableStateOf<List<Species>>(listOf<Species>())
    val counties = mutableStateOf<List<String>>(listOf<String>())
    val parks = mutableStateOf<List<Park>>(listOf<Park>())
    var observations = mutableStateListOf<Observation>()
    val userid = mutableStateOf(0)

    init {
        val db = Room.databaseBuilder(
            application,
            InvasiveDatabase::class.java, "invasive"
        ).build()
        dao = db.invasiveDao()

        val okHttpClient = OkHttpClient.Builder()
            .connectTimeout(30, TimeUnit.SECONDS) // Connect timeout
            .readTimeout(30, TimeUnit.SECONDS)    // Read timeout
            .writeTimeout(30, TimeUnit.SECONDS)   // Write timeout
            .build()

        val retrofit : Retrofit = Retrofit.Builder()
            .addConverterFactory(GsonConverterFactory.create())
            .baseUrl("https://cmsc106.net/invasive/")
            .client(okHttpClient)
            .build()
        restInterface = retrofit.create(InvasiveApiService::class.java)

        loadAllFromDB()

        if(species.value.isEmpty()) {
            loadAllFromServer()
        }
    }

    fun loadAllFromDB() {
        viewModelScope.launch(Dispatchers.IO) {
            val speciesList = dao.loadSpecies()
            val countiesList = dao.loadCounties()
            var parksList : List<Park>
            if(countiesList.isNotEmpty()) {
                parksList = dao.loadCountyParks(county = countiesList[0])
            } else
                parksList = listOf<Park>()
            val observationList = dao.loadObservations()

            withContext(Dispatchers.Main) {
                species.value = speciesList
                observations = observationList.toMutableStateList()
                counties.value = countiesList
                parks.value = parksList
            }
        }
    }

    fun loadAllFromServer() {
        viewModelScope.launch(Dispatchers.IO) {
            try {
                var speciesList = restInterface.getSpecies()
                for(sp in speciesList) {
                    dao.insertSpecies(sp)
                }
                speciesList = dao.loadSpecies()

                var parksList = restInterface.getParks()
                for(pk in parksList) {
                    dao.insertPark(pk)
                }
                val countiesList = dao.loadCounties()
                parksList = dao.loadCountyParks(county = countiesList[0])

                withContext(Dispatchers.Main) {
                    species.value = speciesList
                    counties.value = countiesList
                    parks.value = parksList
                }
            } catch (e:Exception) {
                e.printStackTrace()
            }
        }
    }

    fun loadParksForCounty(county : String) {
        viewModelScope.launch(Dispatchers.IO) {
            val parksList : List<Park>
            if(counties.value.isNotEmpty()) {
                parksList = dao.loadCountyParks(county = county)
            } else
                parksList = listOf<Park>()

            withContext(Dispatchers.Main) {
                parks.value = parksList
            }
        }
    }

    fun newUser(user : User,onSuccess : ()->Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            try {
                val result = restInterface.newUser(user)
                withContext(Dispatchers.Main) {
                    userid.value = result.iduser
                    onSuccess()
                }
            } catch (e:Exception) {
                e.printStackTrace()
            }
        }
    }

    fun login(name : String,password : String,onSuccess : ()->Unit) {
        viewModelScope.launch(Dispatchers.IO) {
            try {
                val result = restInterface.login(name,password)
                withContext(Dispatchers.Main) {
                    userid.value = result.iduser
                    onSuccess()
                }
            } catch (e:Exception) {
                e.printStackTrace()
            }
        }
    }

    fun addObservation(observation : Observation) {
        observations.add(observation)
        viewModelScope.launch(Dispatchers.IO) {
            dao.insertObservation(observation)
        }
    }

    fun uploadObservations() {
        viewModelScope.launch(Dispatchers.IO) {
            try {
                for(obs in observations) {
                    obs.user = userid.value
                    restInterface.postObservation(obs)
                    dao.removeObservation(obs)
                }
                withContext(Dispatchers.Main) {
                    observations.clear()
                }
            } catch (e:Exception) {
                e.printStackTrace()
            }
        }
    }
}